_funcs = None


class DotAccessibleNone:
    __dict__ = None

    def __getattr__(self, k):
        return DotAccessibleNone()


class DotAccessible(dict):
    """
    Dot accessible version of a dict. For syntactic sugar.

    >>> it = DotAccessible({"a": 5})
    >>> it.a
    5
    >>> it.b = 6
    >>> it
    {'a': 5, 'b': 6}
    >>> del it.b
    >>> isinstance(it.b, DotAccessibleNone)
    True
    >>> it
    {'a': 5}
    >>> DotAccessible({"a": 5}, b=1)
    {'a': 5, 'b': 1}
    """

    def __init__(self, *args, **kwargs):
        super(DotAccessible, self).__init__(*args, **kwargs)
        for arg in args:
            if isinstance(arg, dict):
                for k, v in arg.items():
                    self[k] = dotaccessible(v)

        if kwargs:
            for k, v in kwargs.items():
                self[k] = dotaccessible(v)

    def __getattr__(self, k):
        return dotaccessible(
            self.get(k) if not k.startswith("__") else super().__getattr__(k)
        )

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __setitem__(self, key, value):
        super(DotAccessible, self).__setitem__(key, value)
        self.__dict__.update({key: value})

    def __delattr__(self, item):
        self.__delitem__(item)

    def __delitem__(self, key):
        super(DotAccessible, self).__delitem__(key)
        del self.__dict__[key]


def dotaccessible(it):
    if it is None:
        return DotAccessibleNone()
    if isinstance(it, dict):
        return DotAccessible(it)
    return it


def worker_init(funcs):
    """
    initializer for the worker in multiprocessing
    """
    global _funcs
    _funcs = funcs


class JFREMOVED:
    pass


def worker(x):
    """
    worker for multiprocessing
    >>> worker_init([["map", lambda x: x],
    ...              ["update", lambda x: x],
    ...              ["function", lambda x: lambda y: y],
    ...              ["filter", lambda x: x]])
    >>> worker({"a": 1})
    {'a': 1}
    """
    for op, _func in _funcs:
        if op == "map":
            x = _func(dotaccessible(x))
        elif op == "update":
            x.update(_func(dotaccessible(x)))
        elif op == "function":
            x = _func(dotaccessible(x))(dotaccessible(x))
        elif op == "filter":
            test = _func(dotaccessible(x))
            if not test:
                return JFREMOVED
    return x


def dict_updater(_f):
    def _update_dict(x):
        return dict(x, **_f(x))

    return _update_dict


def mymap(fs, arr, processes=1):
    """My mapping function

    Apply functions in fs to items in arr. Also supports multiprocessing.

    """
    if processes > 1:
        from multiprocessing import Pool

        with Pool(
            processes,
            initializer=worker_init,
            initargs=([(op, f) for op, f in fs if op != "function"],),
        ) as pool:
            ret = pool.imap(worker, arr, chunksize=16)
            for op, f in fs:
                if op == "function":
                    ret = f(ret)(map(dotaccessible, ret))
            yield from filter(lambda x: x != JFREMOVED, ret)
    else:
        for op, _f in fs:
            if op == "map":
                arr = map(_f, map(dotaccessible, arr))
            elif op == "update":
                arr = map(dict_updater(_f), map(dotaccessible, arr))
            elif op == "function":
                arr = _f(1)(map(dotaccessible, arr))
            elif op == "filter":
                arr = filter(_f, map(dotaccessible, arr))
        yield from arr


def camel_to_snake(name):
    import re

    name = re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()
    return name


def run_query(query, data, additionals={}, from_file=False, processes=1):
    """
    Run query. This function will utilize global imports if used as a library:

    >>> import hashlib
    >>> list(run_query('.a', [{"a": "521"}, {"a": "643"}]))
    ['521', '643']
    """
    from .query_parser import parse_query
    from . import extra_functions
    import parser

    # query
    queries, imports, import_path, _, _ = parse_query(query, from_file, [], [], False)

    name_alternatives = {
        "first": ["head"],
        "last": ["tail"],
        "firstnlast": ["headntail"],
    }

    # environment
    world = dict(
        {"data": data, "mymap": mymap},
        **{
            camel_to_snake(k): getattr(extra_functions, orig_k)
            for orig_k in dir(extra_functions)
            if orig_k[0] != "_"
            for k in name_alternatives.get(camel_to_snake(orig_k), []) + [orig_k]
        },
    )
    if additionals:
        world.update(additionals)
    else:
        import inspect

        def superglobals():
            _globals = dict(
                inspect.getmembers(inspect.stack()[len(inspect.stack()) - 1][0])
            )["f_globals"]
            return _globals

        world.update({k: v for k, v in superglobals().items() if "_" != k[0]})

    for init in additionals.get("JF_init_codes", []):
        eval(parser.suite(init).compile("myinit.py"), world)

    # process
    return eval(
        parser.expr(f"mymap({queries}, data, {processes})").compile("myquery.py"), world
    )



import click
from .main import jf


@click.command()
@click.option("--processes", default=1, help="Number of processes to use.")
@click.option(
    "--from_file",
    "-f",
    help="read transformation from file. This is also activated for query strings that ends with '.jf'",
    is_flag=True,
)
@click.option("--import", "imports", help="additional imports.", multiple=True)
@click.option(
    "--import_from",
    "import_path",
    help="additional import path.",
    multiple=True,
    default=[],
)
@click.option("--compact", "-c", help="compact output.", is_flag=True)
@click.option("--debug", help="show debug.", is_flag=True)
@click.option("--init", help="run initialization code", multiple=True)
@click.option("--raw", "-r", help="raw output.", is_flag=True)
@click.option(
    "--input",
    "inputfmt",
    help="force input format (json, yaml, excel, csv, ...)",
    default=None,
)
@click.option(
    "--output",
    "output",
    help="output format (json, yaml, excel, csv, ...)",
    default="json",
)
@click.argument("query_and_files", nargs=-1, default=None)
def main(
    processes,
    query_and_files,
    imports,
    import_path,
    from_file,
    compact,
    inputfmt,
    output,
    debug,
    raw,
    init,
):
    return jf(
        processes,
        query_and_files,
        imports,
        import_path,
        from_file,
        compact,
        inputfmt,
        output,
        debug,
        raw,
        init,
    )


if __name__ == "__main__":  # pragma: no coverage
    main()



from abc import ABC, abstractmethod


class JFTransformation(ABC):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def _fn(self, arr):
        """This method must be overwrite by actual transformation methods"""

    def __call__(self, arr):
        return self._fn(arr)



from .meta import JFTransformation
from itertools import islice, chain
from queue import deque


class Flatten(JFTransformation):
    """Yield all subitems of all item

    >>> list(Flatten(lambda x: x["a"])([{"a": [1,2,3], "b": [{"c": 1}], "c": {"d": 1}}]))
    [{'a.0': 1, 'a.1': 2, 'a.2': 3, 'b.0.c': 1, 'c.d': 1}]
    """

    def _flatten(self, it, root=""):
        if not isinstance(it, dict):
            return it
        ret = {}
        for key, val in it.items():
            if isinstance(val, dict):
                for k2, v2 in self._flatten(val, root=key + ".").items():
                    ret[k2] = v2
            elif isinstance(val, list):
                for idx, v2 in enumerate(val):
                    dct2 = self._flatten(v2, root=key + ".%d." % idx)
                    if isinstance(dct2, dict):
                        for k3, v3 in dct2.items():
                            ret[k3] = v3
                    else:
                        ret[key + ".%d" % idx] = dct2
            else:
                ret[root + key] = val
        return ret

    def _fn(self, arr):
        for it in arr:
            yield self._flatten(it)


class JfDel(JFTransformation):
    """Yield all subitems of all item

    >>> list(YieldFrom(lambda x: x["a"])([{"a": [1,2,3]}]))
    [1, 2, 3]
    """

    def _fn(self, arr):
        param = self.args[0](1).split(".")
        for item in arr:
            c = item
            for p in param[:-1]:
                c = getattr(c, p)
            if param[-1] in c:
                c.__delitem__(param[-1])
            yield item


class YieldFrom(JFTransformation):
    """Yield all subitems of all item

    >>> list(YieldFrom(lambda x: x["a"])([{"a": [1,2,3]}]))
    [1, 2, 3]
    """

    def _fn(self, arr):
        for items in arr:
            for val in self.args[0](items):
                yield val


class GroupBy(JFTransformation):
    """Group items by value

    >>> list(GroupBy(lambda x: x["a"])([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{1: [{'a': 1}, {'a': 1}], 2: [{'a': 2}]}]
    """

    def _fn(self, arr):
        ret = {}
        for item in arr:
            val = self.args[0](item)
            if val in ret:
                ret[val].append(item)
            else:
                ret[val] = [item]
        yield ret


class Transpose(JFTransformation):
    """Transpose input
    >>> list(Transpose(lambda x: x["a"])([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': [1, 1, 2]}]
    """

    def _fn(self, X):
        import pandas as pd

        df = pd.DataFrame([dict(x) for x in X])
        yield df.to_dict(orient="list")


class Unique(JFTransformation):
    """Calculate unique according to function

    >>> list(Unique(lambda x: x["a"])([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 1}, {'a': 2}]
    >>> list(Unique()([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 1}, {'a': 2}]
    """

    def _fn(self, X):
        def fun(x):
            return repr(x)

        if len(self.args) > 0:
            fun = self.args[0]

        seen = set()
        for it in X:
            h = hash(fun(it))
            if h in seen:
                continue
            else:
                seen.add(h)
                yield it


class Firstnlast(JFTransformation):
    """
    Show first and last (N) items
    >>> list(Firstnlast()([{"a": 1}, {"a": 1}, {"a": 2}]))
    [[{'a': 1}], [{'a': 2}]]
    >>> list(Firstnlast("1")([{"a": 1}, {"a": 1}, {"a": 2}]))
    [[{'a': 1}], [{'a': 2}]]
    """

    def _fn(self, arr):
        shown = 1
        if len(self.args) == 1:
            shown = self.args[0]
        if callable(shown):
            shown = shown(1)
        if not isinstance(shown, int):
            shown = 1
        return [list(islice(arr, 0, shown)), list(iter(deque(arr, maxlen=shown)))]


class Chain(JFTransformation):
    """
    Show only the first (N) value(s)
    >>> list(Chain()(Firstnlast(lambda x: 1)([{"a": 1}, {"a": 1}, {"a": 2}])))
    [{'a': 1}, {'a': 2}]
    """

    def _fn(self, arr):
        arr = list(arr)
        ret = chain(*arr)
        return ret


class First(JFTransformation):
    """
    Show only the first (N) value(s)
    >>> list(First(lambda x: 1)([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 1}]
    >>> list(First("1")([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 1}]
    """

    def _fn(self, arr):
        shown = 1
        if len(self.args) == 1:
            shown = self.args[0]
        if callable(shown):
            shown = shown(1)
        if not isinstance(shown, int):
            shown = 1
        ret = islice(arr, 0, shown)
        return ret


class Last(JFTransformation):
    """
    Show only the last (N) value(s)
    >>> list(Last(lambda x: 1)([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 2}]
    >>> list(Last("1")([{"a": 1}, {"a": 1}, {"a": 2}]))
    [{'a': 2}]
    """

    def _fn(self, arr):
        shown = 1
        if len(self.args) == 1:
            shown = self.args[0]
        if callable(shown):
            shown = shown(1)
        if not isinstance(shown, int):
            shown = 1
        ret = reversed(list(First(shown)(reversed(arr))))
        return ret


class Sorted(JFTransformation):
    """
    Sort items based on the column value
    >>> list(Sorted(lambda x: x["a"])([{"a": 3}, {"a": 1}, {"a": 2}]))
    [{'a': 1}, {'a': 2}, {'a': 3}]
    """

    def _fn(self, X):
        keyget = None
        if len(self.args) == 1:
            keyget = self.args[0]
        ret = sorted(X, key=keyget, **self.kwargs)
        return ret


class Print(JFTransformation):
    """
    Print (n) values

    This prints n values to the stderr, but passes the data through without changes.

    >>> list(Print(2)([{"a": 3}, {"a": 1}, {"a": 2}]))
    [{'a': 3}, {'a': 1}, {'a': 2}]
    >>> list(Print(lambda x: 2)([{"a": 3}, {"a": 1}, {"a": 2}]))
    [{'a': 3}, {'a': 1}, {'a': 2}]
    """

    def _fn(self, arr):
        import sys
        import json

        n = 1
        if len(self.args) > 0:
            n = self.args[0]
        if callable(n):
            n = n(1)
        arr = list(arr)
        for it in islice(arr, 0, n):
            sys.stderr.write(json.dumps(it) + "\n")
        return arr


def age(datestr):
    """
    Age of a datetime string

    >>> age("1 weeks ago").days
    6
    """
    from datetime import datetime, timezone, timedelta
    from dateparser import parse as parsedate

    try:
        ret = datetime.now() - parsedate(str(datestr))
    except TypeError:
        ret = datetime.now(timezone.utc) - parsedate(str(datestr))
    return ret



def yield_json_and_json_lines(inp):
    """Yield json and json lines

    Split potentially huge json strings into lines or components for low memory data processing.

    Notice: Results are still json strings, so you most likely want to json.loads them.

    """
    from jf import jsonlgen

    return jsonlgen.gen(iter(inp))


class MinimalAdapter:
    """
    >>> a = MinimalAdapter()
    >>> a(iter([b"abcde", b"fghij"])).read(2)
    'ab'
    >>> a.read(2)
    'cd'
    >>> a.read(2)
    'ef'
    """

    def __init__(self):
        self._fip = None
        self._buf = None  # storage of read but unused material, maximum one line

    def __call__(self, fip):
        self._fip = fip  # store for future use
        self._buf = ""
        return self

    def read(self, size):
        if len(self._buf) >= size:
            # enough in buffer from last read, just cut it off and return
            tmp, self._buf = self._buf[:size], self._buf[size:]
            return tmp
        for line in self._fip:
            self._buf += line.decode()
            if len(self._buf) > size:
                break
        else:
            # ran out of lines, return what we have
            tmp, self._buf = self._buf, ""
            return tmp
        tmp, self._buf = self._buf[:size], self._buf[size:]
        return tmp


def get_handler(method, fntype, additionals):
    fnname = f"jf_{fntype}_{method}"
    for mod, it in additionals.items():
        if any([x == fnname for x in dir(it)]):
            return getattr(it, fnname)


def fetch_http(url):
    import requests

    return lambda url: requests.get(url).content


def fetch_https(url):
    import requests

    return lambda url: requests.get(url).content


def fetch_file(fn, f, additionals):
    """
    Fetch file with custom handler

    >>> from io import StringIO
    >>> s = StringIO()
    >>> class fetch_mod:
    ...     def jf_fetch_s3(m):
    ...          return '{"hello": "world"}'
    >>> fetch_file("s3://bucket/key.json", s, {"mod": fetch_mod})
    >>> s.getvalue()
    '{"hello": "world"}'
    """
    from itertools import chain

    proto = fn.split("://")[0]
    try:
        fun = globals()[f"fetch_{proto}"](fn)
    except Exception as ex:
        print(ex)
        fun = get_handler(proto, "fetch", additionals)
    if fun:
        f.write(fun(fn))
        return
    raise NotImplementedError(
        f"I do not know how to fetch {proto}://.\nPlease implement {proto}(url) -> bytes and import a module with it with --import"
    )


def data_input(files=None, additionals={}, inputfmt=None):
    """
    Data input function

    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile() as tmpfile:
    ...     tmpfile.write(b'[{"myconfig": "myvalue"}]') and True
    ...     tmpfile.flush()
    ...     len(list(data_input([tmpfile.name])))
    True
    1
    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmpfile:
    ...     tmpfile.write(b'[{"myconfig": "myvalue"}]') and True
    ...     tmpfile.flush()
    ...     len(list(data_input([tmpfile.name])))
    True
    1
    >>> with tempfile.NamedTemporaryFile(suffix=".csv") as tmpfile:
    ...     tmpfile.write(b"hello,world\\nno,yes\\nno,no\\n") and True
    ...     tmpfile.flush()
    ...     len(list(data_input([tmpfile.name])))
    True
    2
    >>> list(data_input(["nots3://bucket/key.json"], {}))
    Traceback (most recent call last):
    ...
    NotImplementedError: ...
    """
    import fileinput
    import json
    import sys

    pandas_ext = (
        "csv",
        "xlsx",
        "feather",
        "fwf",
        "gbq",
        "hdf",
        "html",
        "orc",
        "parquet",
        "pickle",
        "sas",
        "spss",
        "sql",
        "stata",
        "xml",
    )

    def try_json_loads(it):
        try:
            return json.loads(it)
        except Exception as ex:
            pass

    if not files:
        yield from filter(
            lambda x: x, map(try_json_loads, yield_json_and_json_lines(sys.stdin))
        )
    tmpf = None
    for fn in files:
        inputfmt = fn.split(".")[-1] if inputfmt is None else inputfmt
        inputfmt = inputfmt.split(",", 1)
        inputfmt, inputkwargs = inputfmt[0], inputfmt[1] if len(inputfmt) == 2 else ""
        inputkwargs = (
            dict([it.split("=") for it in inputkwargs.split(",")])
            if len(inputkwargs)
            else {}
        )
        if "://" in fn:
            from tempfile import NamedTemporaryFile

            tmpf = NamedTemporaryFile(suffix="." + fn.split(".")[-1], delete=False)
            fn = fetch_file(fn, tmpf, additionals)
            tmpf.close()
            fn = tmpf.name
        if inputfmt in pandas_ext:
            import pandas

            df = getattr(pandas, f"read_{inputfmt}")(fn, **inputkwargs)
            for it in df.to_dict(orient="records"):
                yield it
            continue
        if inputfmt in ("yml", "yaml"):
            import yaml

            ma = MinimalAdapter()
            with fileinput.input(
                fn, openhook=fileinput.hook_compressed, mode="rb"
            ) as f:
                ret = yaml.safe_load(ma(f))
                if isinstance(ret, list):
                    yield from ret
                else:
                    yield ret
            continue
        if not inputfmt in ("json", "jsonl"):
            fun = get_handler(fn.split(".")[-1], "unserialize", additionals)
            if fun:
                with open(fn, "rb") as f:
                    yield from fun(f)
                    continue
        with fileinput.input(fn, openhook=fileinput.hook_compressed, mode="rb") as f:
            yield from map(
                try_json_loads, yield_json_and_json_lines(map(lambda x: x.decode(), f))
            )
        if tmpf:
            import os

            os.unlink(tmpf.name)
            tmpf = None


def write_bytes(barr):
    import io
    import os
    import sys

    try:
        with os.fdopen(sys.stdout.fileno(), "wb", closefd=False) as stdout:
            stdout.write(barr)
            stdout.flush()
    except io.UnsupportedOperation:
        print("<bytes>")


def save_pandas(alldata, output, _highligh=None):
    import pandas as pd
    from io import BytesIO

    df = None
    try:
        df = pd.DataFrame([dict(it) for it in alldata])
        if output == "xlsx":
            output = "excel"
        res = BytesIO()
        try:
            getattr(df, f"to_{output}")(res)
            try:
                ret = res.getvalue().decode()
                ret = ret if not _highligh else _highligh(ret)
                print(ret)
            except:
                write_bytes(res.getvalue())
        except:
            from io import StringIO

            try:
                ress = StringIO()
                getattr(df, f"to_{output}")(ress)
                ret = ress.getvalue() if not _highligh else _highligh(ress.getvalue())
                print(ret)
            except TypeError:
                try:
                    print(getattr(df, f"to_{output}")())
                except TypeError:
                    import os
                    import tempfile

                    with tempfile.NamedTemporaryFile(delete=False) as tmpfile:
                        tmpfile.close()
                        getattr(df, f"to_{output}")(tmpfile.name)
                        with open(tmpfile.name, "rb") as f:
                            write_bytes(f.read())
                        os.unlink(tmpfile.name)
        return
    except ImportError as err:
        print(f"Failed to import required dependency for {output}")
        print(err)
    except ValueError as err:
        import sys

        sys.stderr.write("Failed to handle {}".format(alldata))
        sys.stderr.write("Failed to handle {}".format(df))
        sys.stderr.write(repr(err))


def get_supported_formats():
    """
    >>> len(get_supported_formats()) > 2
    True
    """
    import pandas as pd

    df = pd.DataFrame([{"a": 1}])
    SUPPORT_EXCLUDE = {
        "timestamp",
        "hdf",
        "sql",
        "period",
        "records",
        "dict",
        "stata",
        "gbq",
    }
    return set(
        [
            it[3:]
            for it in dir(df)
            if it.startswith("to_")
            if not it[3:] in SUPPORT_EXCLUDE
        ]
        + ["json", "jsonl", "yaml", "python", "py"]
    )


def print_results(ret, output, compact=False, raw=False, additionals={}):
    """
    Print array with various formats

    >>> data = [{"a": 1}]
    >>> print_results(data, 'help')
    - clipboard
    - csv
    ...
    >>> print_results(data, 'py', True)
    {'a': 1}
    >>> print_results(data, 'json', True)
    {"a": 1}
    >>> print_results(["hello"], 'json', True, raw=True)
    hello
    >>> print_results(data, 'json', False)
    {
      "a": 1
    }
    >>> print_results(data, 'yaml')
    - a: 1
    <BLANKLINE>
    >>> print_results(data, 'csv')
    ,a
    0,1
    <BLANKLINE>
    >>> print_results(data, 'pickle')
    <bytes>
    >>> class serialize_mod:
    ...     def jf_serialize_msg(m):
    ...          return repr(m)
    >>> print_results(data, 'msg', additionals={"mod": serialize_mod})
    <bytes>
    >>> print_results(data, 'not supported')
    Traceback (most recent call last):
    ...
    NotImplementedError: Cannot output not supported yet. Please consider making a PR!
    """
    import sys
    import json
    from pygments.lexers import get_lexer_by_name
    from pygments import highlight
    from pygments.formatters import TerminalFormatter

    if output == "yaml":
        import yaml
    if output == "help":
        import yaml

        print(yaml.dump(list(sorted(get_supported_formats()))))
        return

    class StructEncoder(json.JSONEncoder):
        """
        Try to convert everything to json

        >>> from datetime import datetime
        >>> import json
        >>> len(json.dumps(datetime.now(), cls=StructEncoder)) > 10
        True
        """

        def default(self, obj):
            try:
                return obj.__dict__
            except AttributeError:
                return obj.__str__()

    _highligh = None
    try:
        formatter = TerminalFormatter()
        lexertype = output if output != "jsonl" else "json"
        output_kwargs = {}
        if not compact:
            output_kwargs["indent"] = 2
        if sys.stdout.isatty():
            lexer = get_lexer_by_name(lexertype, stripall=True)
            _highligh = lambda line: highlight(line, lexer, formatter).rstrip()
    except:
        pass
    ret = iter(ret)
    for line in ret:
        out = line
        if output in ("python", "py"):
            line = repr(line)
        elif output == "yaml":
            line = yaml.dump([dict(line)] if isinstance(line, dict) else line)
        elif output in ("json", "jsonl"):
            line = json.dumps(
                line, ensure_ascii=False, cls=StructEncoder, **output_kwargs
            )
        else:
            alldata = [line] + list(ret)
            fun = get_handler(output, "serialize", additionals)
            if fun:
                ret = fun(alldata)
                write_bytes(ret)
                return
            try:
                return save_pandas(alldata, output, _highligh)
            except Exception as err:
                sys.stderr.write(
                    "Cant produce {}. We only know how to do {} ({})".format(
                        repr(output), get_supported_formats(), repr(err)
                    )
                )
                raise NotImplementedError(
                    f"Cannot output {output} yet. Please consider making a PR!"
                )
        if raw:
            if isinstance(out, str):
                # Strip quotes
                line = line[1:-1]
            if isinstance(out, bytes):
                sys.stdout.buffer.write(line)
            else:
                print(line)
        else:
            print(_highligh(line)) if _highligh else print(line)



from .query_parser import parse_query
from .process import run_query, dotaccessible
from .jfio import data_input, print_results


def jf(
    processes,
    query_and_files,
    imports,
    import_path,
    from_file,
    compact,
    inputfmt,
    output,
    debug,
    raw,
    init,
):
    """Main of the machine

    >>> import tempfile
    >>> import json
    >>> def run_with_data(data, jffn):
    ...     with tempfile.NamedTemporaryFile(suffix='.json') as tmpfile:
    ...         tmpfile.write(json.dumps(data).encode()) and True
    ...         tmpfile.flush()
    ...         jffn(tmpfile.name)
    >>> def jffn(query, *args):
    ...     def _fn(fname):
    ...         ret1 = jf(1, [query, fname], *args)
    ...         ret2 = jf(2, [query, fname], *args)
    ...         assert ret1 == ret2
    ...     return _fn
    >>> run_with_data([{"a": "myvalue"}], jffn(".a", [], [], False, True, None, 'json', False, False, []))
    "myvalue"
    "myvalue"
    >>> run_with_data([{"a": "myvalue", "b": 1}], jffn("{a}", [], [], False, True, None, 'json', False, False, []))
    {"a": "myvalue"}
    {"a": "myvalue"}
    >>> run_with_data([{"a": "myvalue", "b": 1}], jffn("{a, good: .b > 0}", [], [], False, True, None, 'json', False, False, []))
    {"a": "myvalue", "good": true}
    {"a": "myvalue", "good": true}
    >>> run_with_data([{"a": "myvalue"}], jffn("{b: .a}", [], [], False, True, None, 'json', False, False, []))
    {"b": "myvalue"}
    {"b": "myvalue"}
    >>> run_with_data([{"a": 1}, {"a": 2}], jffn("(.a > 1)", [], [], False, True, None, 'json', False, False, []))
    {"a": 2}
    {"a": 2}
    >>> run_with_data([{"a": "myvalue"}], jffn("{hash: hashlib.md5(.a.encode()).hexdigest(), ...}", ["hashlib"], [], False, True, None, 'json', False, False, []))
    {"a": "myvalue", "hash": "d724a7135ce7d2593c25fc5212d4125a"}
    {"a": "myvalue", "hash": "d724a7135ce7d2593c25fc5212d4125a"}
    >>> run_with_data([{"a": "myvalue"}], jffn("{hash: hashlib.md5(.a.encode()).hexdigest(),...}", ["hashlib", "iotools"], ["examples"], False, True, None, 'msg', False, False, []))
    <bytes>
    <bytes>
    >>> run_with_data([{"a": "myvalue"}], jffn("{hash: hashlib.md5(.a.encode()).hexdigest(), c: C, ...}", ["hashlib"], [], False, True, None, 'json', False, False, ["C=5"]))
    {"a": "myvalue", "hash": "d724a7135ce7d2593c25fc5212d4125a", "c": 5}
    {"a": "myvalue", "hash": "d724a7135ce7d2593c25fc5212d4125a", "c": 5}
    """
    import os

    query = "x"
    files = []
    if len(query_and_files) > 0:
        query = query_and_files[0]
    if len(query_and_files) > 1:
        files = query_and_files[1:]

    # input query
    queries, imports, import_path, inputfmt, init = parse_query(
        query,
        from_file,
        imports,
        import_path,
        debug,
        inputfmt=inputfmt,
        init=list(init),
    )

    # environment
    additionals = list(filter(lambda x: is_datafile(x), imports))
    imports = list(filter(lambda x: not is_datafile(x), imports))
    additionals = {
        name(filename): list(map(dotaccessible, data_input([filepath(filename)])))
        for filename in additionals
    }
    additionals = {k: v[0] if len(v) == 1 else v for k, v in additionals.items()}
    additionals["env"] = dotaccessible({k: v for k, v in os.environ.items()})
    if imports:
        import importlib
        import sys
        import os

        sys.path.append(os.path.dirname("."))
        for path in import_path:
            sys.path.append(path)
        additionals.update(
            dict(
                [[name(imp), importlib.import_module(filepath(imp))] for imp in imports]
            )
        )
    additionals["JF_init_codes"] = [parse_query(i, dosplit=False) for i in init]

    # input data
    data = data_input(files, additionals, inputfmt)

    # processing
    ret = run_query(query, data, additionals, from_file, processes)

    # output
    print_results(ret, output, compact, raw, additionals)


def filepath(x):
    return x.split("=")[-1]


def name(x):
    return x.split("=")[0].split(".")[0]


def is_datafile(x):
    return x.endswith(".json")



#!/usr/bin/env python

from os import path
from codecs import open
from setuptools import find_packages, setup, Extension

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

version = "1.0.4"

setup(
    name="jf",
    version=version,
    description="Python jsonl query engine",
    long_description=long_description,
    author="Lasse Hyyrynen",
    author_email="lasse.hyyrynen@pm.me",
    maintainer="Lasse Hyyrynen",
    maintainer_email="lasse.hyyrynen@pm.me",
    license="MIT",
    keywords=["json", "jsonl", "yaml", "jq"],
    download_url="https://github.com/alhoo/jf/archive/%s.tar.gz" % version,
    url="https://github.com/alhoo/jf",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
    packages=["jf"],
    setup_requires=["setuptools>=20.2.2"],
    ext_modules=[
        Extension(
            "jf.jsonlgen",
            sources=["jf/jsonlgen.cc"],
            extra_compile_args=["-std=c++11"],
        ),
    ],
    install_requires=[
        "pygments>=2.4.0",
        "regex>=2016.3.2",
        "python-dateutil>=2.4.1",
        "ruamel.yaml>=0.16.10",
        "dateparser>=0.6.0",
        "pandas>=0.22.0",
        "Click>=7.0.0",
    ],
    tests_require=[
        "nose>=1.3.0",
        "pandas>=0.22.0",
        "openpyxl>=2.5.4",
        "pylint>=1.8.2",
        "pytest",
        "pytest-coverage",
        "pytest-flake8",
    ],
    extras_require={"dev": ["recommonmark", "sphinx", "sphinx-rtd-theme"]},
    test_suite="tests",
    zip_safe=True,
    entry_points={"console_scripts": ["jf=jf.__main__:main"]},
)



def split_query(q):
    """
    Split input query into components
    """
    pos = 0
    start = 0
    level = 0
    is_function = False
    while pos < len(q):
        if q[pos] in """([{'""":
            level += 1
        elif q[pos] in """)]}'""":
            level -= 1
            if level == 0:
                yield withquerytype(q[start : pos + 1], is_function)
                q = q[pos + 1 :]
                pos = 0
                start = 0
                is_function = False
                continue
        elif level == 0:
            if q[pos] in " ," and not is_function:
                start = start + 1
            else:
                is_function = True
        pos += 1
    if len(q):
        yield withquerytype(q[start:], is_function)


def withquerytype(query, is_function=False):
    """
    Parse query type from query component

    >>> withquerytype('{a: 5, ...}')
    ('update', '{a: 5}')
    >>> withquerytype('{a: 5}')
    ('map', '{a: 5}')
    >>> withquerytype('(.a > 5)')
    ('filter', '(.a > 5)')
    >>> withquerytype('count()', True)
    ('function', 'count()')
    >>> withquerytype('sorted(.a)', True)
    ('function', 'sorted(lambda x: (.a))')
    """
    if is_function:
        if query.startswith("unique"):
            query = query[6:]
            return "function", f"unique(lambda x: {query})"
        if query.startswith("sorted"):
            query = query[6:]
            return "function", f"sorted(lambda x: {query})"
        elif query.startswith("yield from "):
            query = query[11:]
            return "function", f"yield_from(lambda x: {query})"
        else:
            if query.startswith("x."):
                return "function", "lambda y: map(lambda x: " + query + ", y)"
            parts = query.split("(", 1)
            if len(parts) == 1:
                return "function", "lambda y: map(lambda x: " + parts[0] + ", y)"
            if len(parts[1]) < 2:
                return "function", "(".join(parts)
            ret = "function", "(lambda x: ".join(parts)
            return ret
    elif query[-1] == ")":
        return "filter", query
    elif query[-6:] == ", ...}":
        return "update", query[:-6] + "}"
    elif query[-5:] == ",...}":
        return "update", query[:-5] + "}"
    return "map", query


def query_convert(query):
    import re

    fixres = [
        [r"([{,] ?)([a-z]+)(?=, |})", r"\1 \2: .\2"],
        [r"\|", r","],
        [r"\n *", r" "],
        [r'([{,] *)([^{} "\[\]\',]+):', r'\1"\2":'],
        [r"^(\.[a-zA-Z])", r"x\1"],
        [r"([ ({])(\.[a-zA-Z])", r"\1x\2"],
        [r'{"([^"]+)": ([^}]+ for ([^ ]+, ?)?\1(, ?[^ ]+)? in)', r"{\1: \2"],
        [r"\bdel x.([^( ]+)", r'jf_del("\1")'],
    ]
    for fixre, sub in fixres:
        query = re.sub(fixre, sub, query)
    return query


def parse_query(
    query,
    from_file=None,
    imports=[],
    import_path=None,
    debug=False,
    dosplit=True,
    inputfmt=None,
    init=[],
):
    """
    Parse user query

    >>> parse_query("{A: .b}, {c: .A, ...}, (.c>1),unique(), yield from .a")
    ('[["map", lambda x: {"A": x.b}], ["update", lambda x: {"c": x.A}], ["filter", lambda x: (x.c>1)], ["function", lambda x: unique(lambda x: ())], ["function", lambda x: yield_from(lambda x: x.a)]]', [], None, None, [])
    >>> parse_query('{timestamps: t.get(f"train/{.audio}"), ...}')
    ('[["update", lambda x: {"timestamps": t.get(f"train/{x.audio}")}]]', [], None, None, [])
    >>> parse_query('{timestamps: t.get(f"train/{.audio}"), ...}', debug=True)
    ('[["update", lambda x: {"timestamps": t.get(f"train/{x.audio}")}]]', [], None, None, [])
    >>> import tempfile
    >>> with tempfile.NamedTemporaryFile(suffix='.jf') as tmpfile:
    ...     tmpfile.write(b'#!/usr/bin/env jf\\n#import hashlib\\n{hash: hashlib.md5(.a).hexdigest(), ...}') and True
    ...     tmpfile.flush()
    ...     print(parse_query(tmpfile.name))
    True
    ('[["update", lambda x: {"hash": hashlib.md5(x.a).hexdigest()}]]', ['hashlib'], None, None, [])
    """
    if from_file or query.endswith(".jf"):
        with open(query, "r") as f:
            query = ""
            firstline = True
            for line in f:
                if firstline:
                    firstline = False
                    if line.startswith("#!"):
                        continue
                if line.startswith("#input"):
                    inputfmt = line.rstrip().split(" ", 1)[1]
                elif line.startswith("#init"):
                    init.append(line.rstrip().split(" ", 1)[1])
                elif line.startswith("#import"):
                    parts = line.rstrip().split(" ")
                    imports = list(imports) + [parts[1]]
                    if len(parts) > 3:
                        if parts[2] == "from":
                            import_path = list(import_path) + parts[3:]
                elif not line.startswith("#") and len(line) > 1:
                    query += line.rstrip()

    if not dosplit:
        return query_convert(query).replace("x.", "lambda x: x.")
    queries = list(split_query(query_convert(query)))
    queries = (
        "["
        + ", ".join([f'["{qtype}", lambda x: {query}]' for qtype, query in queries])
        + "]"
    )
    if debug:
        import sys

        sys.stderr.write(queries + "\n")
    return queries, imports, import_path, inputfmt, init
